/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// AFX.cpp : Defines the entry point for the DLL application.

#include "StdAfx.h"
#include "AFX.h"
#include "MeshGridDecoder.h"
#include "WaveSurf/WaveSurfDec.h"
#include "BBADecoder.h"
#include "FootPrintDecoder.h"
#include "Nodes.h"
#include "ODP.h"
#include "Executiv.h"
#include "ODDec.h"
#include "Except.h"
#include "SAIT_AFX_STREAM_DECODER.H"	// SOWOO 20021212

static char dumpFileName [] = "Stream0.dmp";


/*-------------------------------------------------------------
The exported function that instantiates the AFX decoder
-------------------------------------------------------------*/

void *CreateInstance ()
{
	return (void *) new AFXDecoder;
}

//////////////////////////////////////////////////////////////////////
// AFXDecoder Class
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

AFXDecoder::AFXDecoder()
{
  m_fDump = NULL;
  m_pDecoder = NULL;
}

AFXDecoder::~AFXDecoder()
{
  delete m_pDecoder;
}

/*-------------------------------------------------------------
Parse the data in the specificInfo field of the ES descriptor,
and configure the decoder appropriately.
-------------------------------------------------------------*/
bool AFXDecoder :: Configure (const BYTE *specificInfo, int nInfoLength, AFXConfig &afxConfig)
{
	try
	{
		if (nInfoLength > 0)
		{
			ODCoder odDec (
				NULL, // GetExecutive() -> GetTraceFile (),
				"AFXConfig",
				0, //GetExecutive() -> GetTraceFlags (),
				GetVersion ()
			); //BF Conformance

			odDec.Reset(specificInfo,	nInfoLength);

			afxConfig.SetLength (nInfoLength * 8);
			afxConfig.EncDec (&odDec, DECODE, NULL);
		}
	}
	catch (BifsException)
	{
		//GetExecutive() -> Error (INVALID_BIFS_CONFIG);
		return false;
	}

	if (afxConfig.afxext == NULL_DESC)
	{
		//GetExecutive() -> Error (INVALID_BIFS_CONFIG);
		return false;
	}

	return true;
}

//////////////////////////////////////////////////////////////////////
// Sets the specific decoder
//////////////////////////////////////////////////////////////////////
BOOL AFXDecoder :: Setup (const BYTE * pSpecificInfo, int nSpecificInfoLength, int nESNumber)
{
	ASSERT(nSpecificInfoLength > 0);

	++dumpFileName [6];
	m_fDump = fopen (dumpFileName, "wb");

	AFXConfig afxConfig;
	if (!Configure(pSpecificInfo, nSpecificInfoLength, afxConfig))
		return FALSE;

	// create the objects and
	// call Setup if a tool-specific decoder specific info has to be passed
	switch (afxConfig.afxext->GetTag())
	{
	case 0: // 3D Mesh Coding
		m_pDecoder = new JW3DMCDecoder; // SOWOO 20021210
		decoderType = AFX_3DMC;
		break;
	case 1: // MeshGrid
		m_pDecoder = new MGDecoder;
		m_pDecoder->Setup(afxConfig.afxext);
		decoderType = AFX_MG;
		break;
	case 2:	// WaveletSubdivisionSurface
		m_pDecoder = new WaveSurfDecoder;
		m_pDecoder->Setup(afxConfig.afxext);
		decoderType = AFX_WSS;
		break;
	case 3: // Coordinate Interpolator Compression
		m_pDecoder = new JWCoordinateInterpolatorDecoder; // SOWOO 20021210
		decoderType = AFX_CI;
		break;
	case 4: // Orientation Interpolator Compression
		m_pDecoder = new JWOrientationInterpolatorDecoder; // SOWOO 20021210
		decoderType = AFX_OI;
		break;
	case 5: // Position Interpolator Compression
		m_pDecoder = new JWPositionInterpolatorDecoder; // SOWOO 20021210
		decoderType = AFX_PI;
		break;
	case 6: // Octree Image
		m_pDecoder = new YDIBRDecoder; // SOWOO 20021210
		decoderType = AFX_DIBR;
		break;
	case 7: // BBA
		m_pDecoder = new BBADecoder; // MP
		decoderType = AFX_BBA;
		break;
	// 2004-07-15, Shinjun {
	case 8: // PointTexture Compression
		m_pDecoder = new PointTextureDecoder;
		decoderType = AFX_PT;
		break;
	// 2004-07-15, Shinjun }
	//ory DMLAB 2007 01 08
	case 9:
		m_pDecoder =  new JW3DMCExtensionDecoder;
		decoderType = AFX_3DMCEXT;
		break;
	//~ory DMLAB
	case 10: // FootPrint Based Compression
		m_pDecoder = new FootPrintDecoder;
		m_pDecoder->Setup(afxConfig.afxext);
		decoderType = AFX_FP;
		break;
	default:   
		ASSERT(0);
	}

	return TRUE;
}

//////////////////////////////////////////////////////////////////////
// Find a compatible format and initialize the decoder.
// In this template we just pick up the first format.
//////////////////////////////////////////////////////////////////////
BOOL AFXDecoder :: SetFormat (Capabilities &formatOptions, DecoderParams &decoderParams)
{
	if (formatOptions.GetFormatCapabilityLength () == 0)
		return FALSE;

	decoderParams.m_nStreamNumber = 0;
	decoderParams.m_format = formatOptions.GetFormatCapability (0);
	decoderParams.m_nOptimalAUSize = 256;
	decoderParams.m_nDBSize = 1024;
	decoderParams.m_nCBSize = 100000;
	decoderParams.m_bRequiresFullAUs = TRUE;

  return TRUE;
}

//////////////////////////////////////////////////////////////////////
// Calculate the output size in bytes
//////////////////////////////////////////////////////////////////////
int AFXDecoder :: GetOptimalOutputSize ()
{
  // decoding buffer
  ASSERT(m_pDecoder != NULL);
  return m_pDecoder->GetOptimalOutputSize();
}

//////////////////////////////////////////////////////////////////////
// Decode one unit.
//////////////////////////////////////////////////////////////////////
BOOL AFXDecoder :: Decode (LPBYTE pInput, int nInputLength, LPBYTE pOutput, int &rnOutputLength)
{

	if (rnOutputLength < nInputLength)
	{	rnOutputLength = nInputLength;
		//return FALSE;
	}

#ifdef _DEBUGBID
  // test for the stream
	memcpy (pOutput, pInput, nInputLength);

  FILE *fp = fopen("AFXStream.dmp", "wb");
  if (fp)
  {
    fwrite(pOutput, 1, nInputLength, fp);
    fclose(fp);
  }
#endif

  return m_pDecoder->Decode(pInput, nInputLength, pOutput, rnOutputLength);
}

//////////////////////////////////////////////////////////////////////
// In this decoder, the decoded data always strats at beginning of output
// block
//////////////////////////////////////////////////////////////////////
int AFXDecoder :: GetDataOffset  (const LPBYTE)
{
	return 0;
}

//////////////////////////////////////////////////////////////////////
// In this decoder, attributes are fixed for all units
//////////////////////////////////////////////////////////////////////
DWORD AFXDecoder :: GetAttributes (ATTRIBUTE_CODE, const LPBYTE)
{
	return 0;
}

void AFXDecoder :: Terminate ()
{
	fclose (m_fDump);
  Finish();
}

//////////////////////////////////////////////////////////////////////
// Run() retrieves units from the input stream, decode them, and dispatch
// units to the output stream, till the input stream's end is exceeded.
//////////////////////////////////////////////////////////////////////
void AFXDecoder :: Run ()
{
  // wait for stream, et decode

	LPBYTE pInData, pOutData;
	int nInSize, nOutSize;
	DWORD dwTime;
	BOOL bEOU;

	while (
		IsActive() &&
		(pInData =
			(LPBYTE) GetInputStream () -> FetchNext (nInSize, dwTime, bEOU, TRUE)
		) !=
		END_OF_STREAM
	)
	{
		if (IsActive())
		{	
			if (decoderType == 7) 
				nOutSize = 10000; 
			else nOutSize = GetOptimalOutputSize ();

			pOutData = (LPBYTE) GetOutputStream () -> Allocate (nOutSize, TRUE, FALSE);
			if (pOutData == NULL)
				throw;
			if (pOutData == END_OF_STREAM)
				return;
			fputc (HIBYTE (HIWORD (dwTime)), m_fDump);
			fputc (LOBYTE (HIWORD (dwTime)), m_fDump);
			fputc (HIBYTE (LOWORD (dwTime)), m_fDump);
			fputc (LOBYTE (LOWORD (dwTime)), m_fDump);
			fputc (HIBYTE (HIWORD (nInSize)), m_fDump);
			fputc (LOBYTE (HIWORD (nInSize)), m_fDump);
			fputc (HIBYTE (LOWORD (nInSize)), m_fDump);
			fputc (LOBYTE (LOWORD (nInSize)), m_fDump);
			fwrite (pInData, nInSize, 1, m_fDump);
			while (!Decode (pInData, nInSize, pOutData, nOutSize))
			{	pOutData = (LPBYTE) GetOutputStream () -> Reallocate (nOutSize);
				if (pOutData == NULL)
					throw;
				if (pOutData == END_OF_STREAM)
					return;
			}
			GetOutputStream () -> Dispatch (nOutSize, dwTime);
		}
		GetInputStream () -> Release (TRUE);
	}
	GetOutputStream () -> Close ();
}
